/*
 * Copyright (c) 1993 Addison-Wesley Publishing Company.
 * All rights reserved.  This source code is provided
 * for its instructional value, and is not guaranteed
 * for any particular purpose.
 */
/* mond.c */
#include <sys/types.h>
#include <sys/stropts.h>
#include <sys/sysinfo.h>
#include <sys/poll.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <nlist.h>
#include <signal.h>

#define DELAY	30	/* delay between samples */
#define NFD		16	/* grow pollfd array in NFD units */
#define PIPENAME	"/var/.monpipe"
#define PIPEMODE	(S_IRUSR|S_IWUSR)
#define ELOG(s)	log((s), strerror(errno))
#define FATAL(s)	{ ELOG(s); exit(1); }

static struct pollfd *pfd;
static struct pollfd ppfd;
static int fd, nfd, maxfd, memfd;
static int pipefd[2];
static time_t cpu1[4];
static time_t cpu2[4];
static float tot1, tot2;
static char *kmem = "/dev/kmem";
static char *adotout = "/unix";
static struct nlist list[] = {
	{ "sysinfo", 0, 0, 0, 0, 0 }, 
	{ "", 0, 0, 0, 0, 0 }
};

static void addclient(void), rmvclient(int);
static int moreclients(void);
static void sigpoll(int), sigalrm(int);
static void gettimes(uchar_t *, uchar_t *, uchar_t *,
	uchar_t *);

extern void log(const char *, ...);
extern void daemonize(const char *);

void
main()
{
	struct stat sbuf;
	struct sigaction sa;

	/*
	 * Become a daemon.
	 */
	daemonize("mond");

	/*
	 * Find the kernel symbol we need.
	 */
	if (nlist(adotout, list) < 0)
		FATAL("cannot get namelist: %s");

	/*
	 * Open the kernel memory image and seek to
	 * the symbol's address.
	 */
	if ((memfd = open(kmem, O_RDONLY)) < 0)
		FATAL("cannot open kernel memory: %s");
	if (lseek(memfd, list[0].n_value, SEEK_SET) == -1)
		FATAL("lseek failed: %s");

	/*
	 * Read only the first four entries in the structure.
	 */
	if (read(fd, &cpu1, sizeof(cpu1)) != sizeof(cpu1))
		FATAL("read of kernel memory failed: %s");

	/*
	 * Calculate the total CPU time.
	 */
	tot1 = cpu1[CPU_USER] + cpu1[CPU_KERNEL] +
	  cpu1[CPU_WAIT] + cpu1[CPU_IDLE];

	/*
	 * If the rendezvous point doesn't exist, create
	 * it.  Otherwise, make sure that another copy of
	 * mond isn't already running.
	 */
	if (stat(PIPENAME, &sbuf) < 0) {
		close(creat(PIPENAME, PIPEMODE));
	} else {
		if (S_ISFIFO(sbuf.st_mode)) {
			log("mon process already running");
			exit(1);
		}
	}

	/*
	 * Allocate the initial pollfd array.
	 */
	maxfd = NFD;
	pfd = malloc(maxfd * sizeof(struct pollfd));
	if (pfd == NULL)
		FATAL("cannot allocate pollfd array: %s");
	/*
	 * Create the pipe for the rendezvous point.
	 */
	if (pipe(pipefd) < 0)
		FATAL("cannot create pipe: %s");

	/*
	 * Save one end of the pipe for receiving poll events
	 * and mount the other end in the file system.
	 */
	ppfd.fd = pipefd[0];
	ppfd.events = POLLIN;
	if (fattach(pipefd[1], PIPENAME) < 0) {
		log("cannot attach pipe to %s: %s", PIPENAME,
		  strerror(errno));
		exit(1);
	}

	/*
	 * Install signal handlers for SIGPOLL and SIGALRM.
	 * Block each signal inside the handlers.
	 */
	sigemptyset(&sa.sa_mask);
	sigaddset(&sa.sa_mask, SIGPOLL);
	sigaddset(&sa.sa_mask, SIGALRM);
	sa.sa_handler = sigalrm;
	sa.sa_flags = 0;
	if (sigaction(SIGALRM, &sa, NULL) < 0)
		FATAL("cannot install handler for SIGALRM: %s");
	sa.sa_handler = sigpoll;
	if (sigaction(SIGPOLL, &sa, NULL) < 0)
		FATAL("cannot install handler for SIGPOLL: %s");

	/*
	 * Register to receive SIGPOLL from the pipe.
	 */
	if (ioctl(pipefd[0], I_SETSIG, S_INPUT) < 0)
		FATAL("cannot I_SETSIG on pipe: %s");

	/*
	 * It's not pretty, but all processing is
	 * done in the signal handlers.  This is the
	 * only place where SIGPOLL and SIGALRM can
	 * interrupt the processing.
	 */
	for (;;) {
		pause();
	}
}

static void
addclient(void)
{
	struct strrecvfd recv;
	int fl;

	/*
	 * Receive a file descriptor from the listener.
	 * If it fails, flush the read side of the pipe.
	 */
	if (ioctl(pipefd[0], I_RECVFD, &recv) < 0) {
		ioctl(pipefd[0], I_FLUSH, FLUSHR);
		return;
	}

	/*
	 * Register to receive SIGPOLL for input events
	 * on the stream.
	 */
	if (ioctl(recv.fd, I_SETSIG, S_INPUT) < 0) {
		ELOG("cannot I_SETSIG on fd: %s");
		close(recv.fd);
		return;
	}

	/*
	 * Get the file flags and place the stream in
	 * nonblocking mode.
	 */
	if ((fl = fcntl(recv.fd, F_GETFL, 0)) < 0) {
		ELOG("cannot get file flags on fd: %s");
		close(recv.fd);
		return;
	}
	fl = (fl & ~O_NDELAY) | O_NONBLOCK;
	if (fcntl(recv.fd, F_SETFL, fl) < 0) {
		ELOG("cannot set file flags on fd: %s");
		close(recv.fd);
		return;
	}

	/*
	 * If no entries are available in the array, try
	 * to extend the array.
	 */
	if ((nfd == maxfd) && (moreclients() < 0)) {
		ELOG("cannot grow pollfd array: %s");
		close(recv.fd);
		return;
	}

	/*
	 * Add the file descriptor to the pollfd array and
	 * increment the number of array entries.  If this
	 * is the only entry in the array, set an alarm.
	 */
	pfd[nfd].fd = recv.fd;
	pfd[nfd].events = POLLIN;
	nfd++;
	if (nfd == 1)
		alarm(DELAY);
}

static int
moreclients(void)
{
	struct pollfd *newpfd;
	int n;

	/*
	 * Calculate the new size of the array and
	 * try to allocate new memory.
	 */
	n = maxfd + NFD;
	newpfd = realloc(pfd, n*sizeof(struct pollfd));
	if (newpfd == NULL)
		return(-1);

	/*
	 * Save the new address of the pollfd array
	 * and the new maximum number of entries.
	 */
	pfd = newpfd;
	maxfd = n;
	return(0);
}

static void
rmvclient(int idx)
{
	/*
	 * Generate a disconnect, just to be sure
	 * the connection gets cleaned up.
	 */
	t_snddis(pfd[idx].fd, NULL);

	/*
	 * Close the network connection.
	 */
	t_close(pfd[idx].fd);

	/*
	 * Copy the last element in the array into
	 * the position of the defunct connection and
	 * decrease the count of the number of entries
	 * in the array.
	 */
	pfd[idx].fd = pfd[nfd-1].fd;
	nfd--;
}

static void
sigalrm(int sig)
{
	int i;
	uchar_t cusr, csys, cwio, cidle;

	/*
	 * No clients -- do nothing.
	 */
	if (nfd == 0)
		return;

	/*
	 * Get the system's time percentages.
	 */
	gettimes(&cusr, &csys, &cwio, &cidle);

	/*
	 * Send each percentage to the client.
	 */
	for (i = 0; i < nfd; i++) {
		write(pfd[i].fd, &cusr, 1);
		write(pfd[i].fd, &csys, 1);
		write(pfd[i].fd, &cwio, 1);
		write(pfd[i].fd, &cidle, 1);
	}

	/*
	 * Reinstall the alarm.
	 */
	alarm(DELAY);
}

static void
sigpoll(int sig)
{
	int i;

	/*
	 * Service the listener as long as it is sending
	 * us file descriptors.
	 */
	while (poll(&ppfd, 1, 0) == 1) {
		if (ppfd.revents == POLLIN) {
			addclient();
		} else {
			log("poll on pipe, revents %d", ppfd.revents);
			exit(1);
		}
	}

	/*
	 * Now check if any of the clients have gone away.
	 * rmvclient will pack the array by shifting the
	 * last valid entry into a deleted slot, so we need
	 * to check entry "i" again if we remove an entry.
	 */
	while (poll(pfd, nfd, 0) > 0) {
		for (i = 0; i < nfd; i++)
			if (pfd[i].revents != 0)
				rmvclient(i--);
	}
}

static void
gettimes(uchar_t *up, uchar_t *sp, uchar_t *wp, uchar_t *ip)
{
	float deltat;
	int usr, sys, wio, idle;

	/*
	 * Seek to the beginning of the sysinfo structure.
	 */
	if (lseek(memfd, list[0].n_value, SEEK_SET) == -1)
		FATAL("lseek failed: %s");

	/*
	 * Read only the first four entries in the structure.
	 */
	if (read(memfd, &cpu2, sizeof(cpu2)) != sizeof(cpu2))
		FATAL("read of kernel memory failed: %s");

	/*
	 * Calculate the new total CPU time.
	 */
	tot2 = cpu2[CPU_USER] + cpu2[CPU_KERNEL] +
	  cpu2[CPU_WAIT] + cpu2[CPU_IDLE];

	/*
	 * Calculate the difference between the current
	 * and previous totals.
	 */
	deltat = tot2 - tot1;

	/*
	 * Calculate the percentages, rounding up and
	 * converting to integers.
	 */
	usr = (cpu2[CPU_USER]-cpu1[CPU_USER])/deltat*100.+
	  0.5;
	sys = (cpu2[CPU_KERNEL]-cpu1[CPU_KERNEL])/deltat*100.+
	  0.5;
	wio = (cpu2[CPU_WAIT]-cpu1[CPU_WAIT])/deltat*100.+
	  0.5;
	idle = (cpu2[CPU_IDLE]-cpu1[CPU_IDLE])/deltat*100.+
	  0.5;

	/*
	 * Save the current values in the previous slots.
	 */
	tot1 = tot2;
	cpu1[CPU_USER] = cpu2[CPU_USER];
	cpu1[CPU_KERNEL] = cpu2[CPU_KERNEL];
	cpu1[CPU_WAIT] = cpu2[CPU_WAIT];
	cpu1[CPU_IDLE] = cpu2[CPU_IDLE];

	/*
	 * Return the percentages.
	 */
	*up = (uchar_t)usr;
	*sp = (uchar_t)sys;
	*wp = (uchar_t)wio;
	*ip = (uchar_t)idle;
}
